﻿/***********************************************************************

						 -- Oyayusby Utility --

------------------------------------------------------------------------

	ﾒｲﾝｳｨﾝﾄﾞｳ

	Copyright (c) 2018 Kiyoto.
	This software is distributed under the MIT license.

***********************************************************************/

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using oyayusby_util_lib;


namespace oyayusby_util
{
	public partial class Main : Form
	{
		/* 本来のｶﾅﾛｯｸ取得用(IsKeyLockedで取ろうとすると例外になるので) */
		[DllImport("User32.dll")]
		private static extern short GetKeyState(int nVirtKey);

		/* このｲﾝｽﾀﾝｽが生きている間はﾒｯｾｰｼﾞﾌｯｸが差し込まれた状態 */
		public MSG_HOOK msg_hook;
		
		/* 64bit版では、32bit版ｲﾝｽﾀﾝｽに処理を集約させる(32bitｲﾝｽﾀﾝｽがいれば、そっちにﾒｯｾｰｼﾞを丸投げ) */
		[DllImport("User32.dll")]
		public static extern IntPtr FindWindow(string lpClassName, string lpWindowName);
		[DllImport("User32.dll")]
		public static extern int PostMessage(IntPtr hWnd, uint Msg, IntPtr wParam, IntPtr lParam);
		public IntPtr hwnd32;				/* ﾒｯｾｰｼﾞ送信先 */
		public string find_window_title;	/* ﾒｯｾｰｼﾞ送信先の捜索用ﾀｲﾄﾙ */

		/* 最後に設定したLED点灯状況 */
		public bool is_locked_kana;			/* このｱﾌﾟﾘで設定しているｶﾅﾛｯｸ */
		public bool is_locked_real_kana;	/* ﾘｱﾙｶﾞﾁなｶﾅﾛｯｸ */
		public bool is_locked_scroll;
		public bool is_locked_num;
		public bool is_locked_caps;

		public Main()
		{
			is_locked_kana = false;
			InitializeComponent();

			ctxmenu_show.Font = new Font(ctxmenu_show.Font, FontStyle.Bold);
			notify.Icon = new Icon(Icon, new Size(16, 16));
			notify.Visible = true;

			/* ﾋﾞﾙﾄﾞ種別に応じた文字列を設定 */
			{
				string target = Environment.Is64BitProcess ? "(x64)" : "(x86)";
				find_window_title = Text + "(x86)";
				Text += target;
				notify.Text += target;
			}
		}

		private void Main_Load(object sender, EventArgs e)
		{
			msg_hook = new MSG_HOOK(Handle);
		}

		protected bool is_logging() => chk_logging.Checked;
		protected void put_log(string format, params object[] args)
		{
			if (txt_log.Text.Length >= 4096)
			{
				txt_log.Text = "";
			}
			txt_log.Text += String.Format(format, args);
			txt_log.Select(txt_log.Text.Length, 0);
			txt_log.ScrollToCaret();
		}

		protected override void WndProc(ref Message m)
		{
			if (m.Msg == MSG_HOOK.message_id)
			{
				if (Environment.Is64BitProcess && (hwnd32 = (hwnd32 == IntPtr.Zero ? FindWindow(null, find_window_title) : hwnd32)) != IntPtr.Zero)
				{
					if (is_logging()) put_log("Through 32bit: conv={0:X8}, pid={1:d}.\r\n", (uint)m.WParam, (uint)m.LParam);
					if (PostMessage(hwnd32, (uint)m.Msg, m.WParam, m.LParam) != 0)
					{
						return;
					}
					if (is_logging()) put_log("...fail\r\n");
					hwnd32 = IntPtr.Zero;
				}
				
				{
					bool is_locked_kana_new      = is_locked_kana;
					bool is_locked_real_kana_new = ((GetKeyState((int)Keys.KanaMode) & 1) != 0)/* 本来のｶﾅﾛｯｸ(但しﾏﾄﾓに管理されていない) */ ;
					bool is_locked_scroll_new    = IsKeyLocked(Keys.Scroll);
					bool is_locked_num_new       = IsKeyLocked(Keys.NumLock);
					bool is_locked_caps_new      = IsKeyLocked(Keys.Capital);

					if (is_logging()) put_log("Msg: conv={0:X8}, pid={1:d}.\r\n", (uint)m.WParam, (uint)m.LParam);

					/* 入力ﾓｰﾄﾞ通知が来た場合は新しい値に更新 */
					/* @memo ｼｽﾃﾑ側でLEDの値を更新する際、ｶﾅﾛｯｸLEDの値 = ｼｽﾃﾑが設定した最終値で上書きされてしまう */
					/*       これを回避する為、LEDが更新され得る場面で前回ｶﾅﾛｯｸ値が含まれるLED設定を発行し直す必要がある */
					/*       上記場面や、状態取得に失敗した場合は、下記を通らないWParamのﾒｯｾｰｼﾞが飛んでくる */
					if ((ulong)m.WParam < 0x10000u){
						is_locked_kana_new = 
							((uint)m.WParam & 0x8000) != 0/* IME ON */ && 
							((uint)m.WParam & 0x0010) == 0/* 非ﾛｰﾏ字 */ && 
							((uint)m.WParam & 0x0001) != 0/* かな / ｶﾅ */;
					}
				
					/* LED設定値が変化した場合のみ(同じ内容の通知が重複して来ることがあるので) */
					if (
						is_locked_kana_new      != is_locked_kana      ||
						is_locked_real_kana_new != is_locked_real_kana ||
						is_locked_scroll_new    != is_locked_scroll    ||
						is_locked_num_new       != is_locked_num       ||
						is_locked_caps_new      != is_locked_caps
					){
						if (is_logging()) put_log("Changed: k:{0}, rk:{1}, s:{2}, n:{3}, c:{4}.\r\n",
							is_locked_kana_new      != is_locked_kana,
							is_locked_real_kana_new != is_locked_real_kana,
							is_locked_scroll_new    != is_locked_scroll,
							is_locked_num_new       != is_locked_num,
							is_locked_caps_new      != is_locked_caps);

						/* ｷｰﾎﾞｰﾄﾞがいつ脱着されても良いように、毎度ｲﾝｽﾀﾝｽを作り直す */
						using (var kbd_enum = new KBD_ENUM())
						foreach (var kbd in kbd_enum.kbd)
						{
							#if false
							{
								/* test: PS/2接続の場合のﾓｰﾄﾞ同期 */
								/* → 「x86版 + 初代親指の友 + USB接続」位しか有効なｼﾁｭｴｰｼｮﾝが無いので、実装却下 */
								/*     x86版だと、PS/2ﾎﾟｰﾄ宛なら正しく通知が来ていたので、この処理は不要だし。 */
								/*     x64版だとi8042prt.sysがIME状態設定を受け付けるｺｰﾄﾞがifdefで消されてたので、使える相手が居ないし。 */
								if (!kbd.set_ime_mode(((uint)m.WParam & 0x8000) != 0 ? 1u : 0u, ((uint)m.WParam & 0x7FFF)))
								{
									textBox2.Text += "...fail.\r\n";
								}
							}
							#endif
						
							/* LED設定値の通知 */
							if (!kbd.set_led(is_locked_scroll_new, is_locked_num_new, is_locked_caps_new, is_locked_kana_new)){
								if (is_logging()) put_log("...fail @ {0}\r\n", kbd.path);
							}
						}

						is_locked_kana      = is_locked_kana_new;
						is_locked_real_kana = is_locked_real_kana_new;
						is_locked_scroll    = is_locked_scroll_new;
						is_locked_num       = is_locked_num_new;
						is_locked_caps      = is_locked_caps_new;
					}
				}
				return;
			}
			base.WndProc(ref m);
		}

		private void cmd_clear_Click(object sender, EventArgs e)
		{
			txt_log.Text = "";
		}

		private void ctxmenu_show_Click(object sender, EventArgs e)
		{
			Visible = true;
			if (WindowState == FormWindowState.Minimized)
			{
				WindowState = FormWindowState.Normal;
			}
			Activate();
		}

		private void ctxmenu_quit_Click(object sender, EventArgs e)
		{
			Close();
		}

		private void Main_ClientSizeChanged(object sender, EventArgs e)
		{
			if (WindowState == FormWindowState.Minimized)
			{
				Visible = false;
			}
		}

		private void Main_FormClosing(object sender, FormClosingEventArgs e)
		{
			notify.Visible = false;
		}
	}
}
